import websockets
from cubevis.utils import is_colab

def create_ws_server(callback, ip_address, port):
    """
    Uniform wrapper for creating a WebSocket server.
    """
    if is_colab( ):
        print( f"websocket startup: {ip_address}/{port} (bind IP 0.0.0.0)" )
        return websockets.serve( callback, "0.0.0.0", port, origins=None )
    else:
        return websockets.serve( callback, ip_address, port )



#from cubevis.bokeh.state import initialize_bokeh
#initialize_bokeh( "../../../casalib/dist/esbuild/casalib-v0.0.1.min.js",
#                  "../../../casaguijs/dist/casaguijs.js" )                        ### local build
#initialize_bokeh( "../../../casalib/dist/esbuild/casalib-v0.0.2.min.js" )
#
#initialize_bokeh( bokeh='../../../casaguijs/dist/casaguijs.js' )                  ### local gui/js build + standard bokeh independent library

import os
import ssl
import certifi
import asyncio
import urllib
import tarfile

from cubevis import iclean

##
## demo measurement set to use
##
ms_path = 'twhya_selfcal_contsub.ms'
##
## where to fetch the demo measurement set
##
ms_url = "https://casa.nrao.edu/download/devel/casavis/data/twhya_selfcal_contsub_ms.tar.gz"
##
## output image file name
##
img =  'twhya_n2hp'

if not os.path.isdir(ms_path):
    try:
        context = ssl.create_default_context(cafile=certifi.where())
        tstream = urllib.request.urlopen(ms_url, context=context, timeout=400)
        tar = tarfile.open(fileobj=tstream, mode="r:gz")
        tar.extractall( )
    except urllib.error.URLError:
        print("Failed to open connection to "+ms_url)
        raise

if not os.path.isdir(ms_path):
    raise  RuntimeError("Failed to fetch measurement set")


print("Result = " + str( repr( iclean( vis=ms_path, imagename=img,
                                       field = '5',
                                       spw = '0',
                                       specmode = 'cube',
                                       perchanweightdensity=True,
                                       nchan = 15,
                                       start = '0.0km/s',
                                       width = '0.5km/s',
                                       outframe = 'LSRK',
                                       restfreq = '372.67249GHz',    ### N2H+ rest frequency
                                       deconvolver= 'hogbom',
                                       gridder = 'standard',
                                       imsize = [500, 500],
                                       cell = '0.1arcsec',
                                       weighting = 'briggsbwtaper',
                                       robust = 0.5,
                                       restoringbeam='common',
                                       niter=100 ) ) ) )



import os
import sys
import importlib.util
from os.path import join, exists, realpath

def find_pkg( name, extra_dirs=[] ):
    """
    Returns list of paths which contain the ``name`` pkg. It is assumed that the
    ``name`` will be the name of a directory (containing the package) or that
    ``name.py(c)`` will be a file that exists. This does not attempt to import the
    package.

    Parameters
    ----------
    name: string
        the name of the package in python "dot" notation to be searched for
    extra_dirs: list of strings
        directories to be searched in addition to those included in ``sys.path``

    Returns
    -------
    list of import specifications
        specifications of ``name`` found by search ``sys.path`` and ``extra_dirs``
    """
    result = [ ]
    filename = (name.split('.'))
    for dir in sys.path[:] + extra_dirs:
        path = join( dir, *filename )
        if exists(path):
            spec = importlib.util.spec_from_file_location(name,path)
            if spec:
                result.append(spec)
        else:
            path = f"{path}.py"
            if exists(path):
                spec = importlib.util.spec_from_file_location(name,path)
                if spec:
                    result.append(spec)
            else:
                path = f"{path}c"
                if exists(path):
                    spec = importlib.util.spec_from_file_location(name,path)
                    if spec:
                        result.append(spec)
    return result

def load_pkg( spec ):
    """
    Loads and returns the module specified by ``spec`` (e.g. as returned from ``find_pkg``)
    If ``spec.name`` does not already exist in ``sys.modules``, it is added.

    Parameters
    ----------
    spec: ModuleSpec
        the module to be loaded

    Returns
    -------
    module
        the loaded module
    """

    import _frozen_importlib
    if not isinstance(spec,_frozen_importlib.ModuleSpec):
        raise ImportError(f"Expected ModuleSpec instead of {type(spec)}")

    pkg = importlib.util.module_from_spec(spec)

    try:
        spec.loader.exec_module(pkg)
    except:
        raise ImportError(f"Could not load {spec.origin}")

    if spec.name not in sys.modules:
        sys.modules[spec.name] = pkg

    return pkg



########################################################################
#
# Copyright (C) 2023
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''Implementation of ``DragTool`` class which provides a events or
callbacks in response to dragging gestures on a figure. Unlike the
PanTool all screen points use the plot's frame of reference instead
of the figure's frame of reference, i.e. sx=0, sy=0 is the lower
left corner of the plot.'''

from bokeh.models import Drag, Tool
from bokeh.core.properties import Instance, Nullable
from bokeh.models.callbacks import Callback
from os.path import join,dirname
from pathlib import Path
from .. import BokehInit

class DragTool(Drag,BokehInit):
    '''Tool that emits press and pressup events to signal start and end of drag'''

    start = Nullable(Instance(Callback), help="""
    JavaScript to be executed when dragging starts. If no start callback has been provided
    a PanStart event will be triggered when dragging starts.
    """)
    move = Nullable(Instance(Callback), help="""
    JavaScript to be executed as dragging progresses. If no move callback has been provided
    a Pan event will be triggered when dragging progresses.
    """)
    end = Nullable(Instance(Callback), help="""
    JavaScript to be executed when dragging ends. If no end callback has been provided
    a PanEnd event will be triggered when dragging stops.
    """)

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        ### neither svg files nor SVGIcon are supported by the icon parameter to a tool
        super().__init__(*args, icon=Path(join(dirname(dirname(dirname(__file__))),'__icons__','drag.png')), **kwargs )

Tool.register_alias("drag", lambda: DragTool())
